#include "StdAfx.h"
#include "FileUploadController.h"
#include "afxinet.h"
#include "json.h"
#include "MemoSendDlg.h"
#include <sstream>
#include "SendMemoController.h"

FileUploadController::FileUploadController(CWnd *pProgressWnd, SendMemoController* pController)
: m_pProgressWnd(pProgressWnd)
, m_pSendMemoController(pController)
, m_maxFileIndex(1)
{
}

FileUploadController::~FileUploadController(void)
{
	std::vector<ThreadInfo>::const_iterator threadIterator = m_threadVector.begin();

	for (; threadIterator != m_threadVector.end(); ++threadIterator)
	{
		ThreadInfo threadInfo = *threadIterator;
		if (::WaitForSingleObject(threadInfo.pUploadThread, 0) != WAIT_OBJECT_0)
		{
			threadInfo.pStopEvent->SetEvent();
			if (::WaitForSingleObject(threadInfo.pUploadThread, 1000) != WAIT_OBJECT_0)
			{
				::TerminateThread(threadInfo.pUploadThread->m_hThread, 0);
			}
		}

		delete threadInfo.pStopEvent;
		delete threadInfo.pUploadThread;
	}
}

UINT FileUploadController::_UploadThread( LPVOID pParam )
{
	
	TRACE("%s %d\n", __FUNCTION__, __LINE__);
	UploadFile::ThreadData* data = (UploadFile::ThreadData*) pParam;

	CEvent *pEvent = data->pStopEvent;
	CString strFilePath = data->filePath;

	if (::WaitForSingleObject(pEvent->m_hObject, 0) == WAIT_TIMEOUT)
	{
		UploadFile uploader;
		if (uploader.FileSend(data) == FALSE)
		{
			CString strMsg = _T("");
			strMsg.Format(IDS_STRING160, strFilePath);
			NMessageBox(AfxGetMainWnd(), strMsg);
		}
	}

	delete data;

	return 0;
}

void FileUploadController::Upload( const CString& strPath)
{

	CEvent* pEvent = new CEvent(FALSE, FALSE);
	if (NULL == pEvent)
		return;

	UploadFile::ThreadData* pThreadData = new UploadFile::ThreadData(strPath, m_pProgressWnd, pEvent, m_maxFileIndex, m_attachID);
	if (NULL == pThreadData)
	{
		delete pEvent;
		return;
	}

	CWinThread* pThread = AfxBeginThread(_UploadThread, pThreadData, THREAD_PRIORITY_NORMAL, 0, CREATE_SUSPENDED);
	if (pThread)
	{
		pThread->m_bAutoDelete = false;

		bool bStart = false;
		std::vector<ThreadInfo>::const_reverse_iterator threadIterator = m_threadVector.rbegin();
		if (threadIterator != m_threadVector.rend())
		{
			ThreadInfo threadInfo = *threadIterator;

			if (::WaitForSingleObject(threadInfo.pUploadThread->m_hThread, 0) == WAIT_OBJECT_0)
			{
				bStart = true;
			}

		}
		else
		{
			bStart = true;
		}
		if (bStart)
		{
			TRACE("%s %d %d\n", __FUNCTION__, __LINE__, m_maxFileIndex);
			m_pSendMemoController->DisableSend();
				
			pThread->ResumeThread();
		}

		m_threadVector.push_back(ThreadInfo(pThread, pEvent, m_maxFileIndex));
		m_maxFileIndex++;
	}
	else
	{
		ASSERT(FALSE);
	}
}

bool FileUploadController::StartNextThread(int fileAttachIndex)
{

	bool bReturn = false;
	using namespace std;

	TRACE("%s %d\n", __FUNCTION__, __LINE__);

	vector<ThreadInfo>::iterator threadIterator = find(m_threadVector.begin(), m_threadVector.end(), ThreadInfo(NULL, NULL, fileAttachIndex));
	if (threadIterator != m_threadVector.end())
	{
		advance(threadIterator, 1);
		if (threadIterator != m_threadVector.end())
		{
			TRACE("%s %d\n", __FUNCTION__, __LINE__);
			ThreadInfo threadInfo = *threadIterator;
			threadInfo.pUploadThread->ResumeThread();
			bReturn = true;
		}
	}
	else
	{
		ASSERT(FALSE);
	}

	return bReturn;
}

bool FileUploadController::Delete( int fileAttachIndex, trid_t *ptrid )
{
//	GT_FLOG(L"file upload controller %d", fileAttachIndex);
	bool bReturn = true;

	std::vector<ThreadInfo>::const_iterator threadIterator = std::find(m_threadVector.begin(), m_threadVector.end(), ThreadInfo(NULL, NULL, fileAttachIndex));
	if (threadIterator == m_threadVector.end())
	{
		g_pMessengerActor->GetWASActor()->RequestDelAttachedFile(m_attachID.c_str(), fileAttachIndex, ptrid);
	}
	else
	{
		ThreadInfo threadInfo = *threadIterator;
		DWORD dwWaitState = ::WaitForSingleObject(threadInfo.pUploadThread->m_hThread, 0);
		if ( dwWaitState == WAIT_OBJECT_0)
		{
			TRACE("%s %d\n", __FUNCTION__, __LINE__);
			g_pMessengerActor->GetWASActor()->RequestDelAttachedFile(m_attachID.c_str(), threadInfo.serverFileIndex, ptrid);
		}
		else
		{
			DWORD dwResult = threadInfo.pUploadThread->ResumeThread();
			threadInfo.pStopEvent->SetEvent();
			TRACE("%s%d\n", __FUNCTION__, __LINE__);
			dwWaitState = ::WaitForSingleObject(threadInfo.pUploadThread->m_hThread, 10000L);
			if (dwWaitState != WAIT_OBJECT_0)
			{
				TRACE("%s%d\n", __FUNCTION__, __LINE__);
				ASSERT(FALSE);
				::TerminateThread(threadInfo.pUploadThread->m_hThread, 0);
			}

			if (dwResult == 0)
			{
				// ༮ UI󿡼  ؾ Ѵ.
				if (m_pProgressWnd && m_pProgressWnd->GetSafeHwnd())
					m_pProgressWnd->PostMessage(UM_MEMO_SEND_LIST_DELETE, (WPARAM)fileAttachIndex, 0);
				
				
				StartNextThread(fileAttachIndex);
			}
			else
			{
				if (m_pProgressWnd && m_pProgressWnd->GetSafeHwnd())
					m_pProgressWnd->PostMessage(UM_MEMO_SEND_LIST_DELETE, (WPARAM)fileAttachIndex, 0);
			}
		}

		delete threadInfo.pStopEvent;
		delete threadInfo.pUploadThread;

		m_threadVector.erase(threadIterator);
	}

	return bReturn;
}

/*!
*  @fn CString MakePreFileData()
*  @brief     ƼƮ ʱ  ͸  κ
*  @return    ʱ  
*  @param     CString & strBoundary
*  @param     CString & strFileName
*  @author	  yangjae.park(yangjae.park@nhn.com) 
*  @date      2012/02/23
*  @warning   
*/
CString UploadFile::MakePreFileData(const CString& strBoundary, const CString& strFileName) 
{ 
	CString strFormat; 
	CString strData; 
	strFormat = _T("\r\n"); 
	strFormat += _T("\r\n"); 
	strFormat += _T("--%s"); 
	strFormat += _T("\r\n"); 
	strFormat += _T("Content-Disposition: form-data; name=\"file\"; filename=\"%s\"\r\n"); 
	strFormat += _T("\r\n"); 
	strData.Format(strFormat, strBoundary, strFileName/*, strBoundary*/); 
	return strData; 
} 


/*!
*  @fn CString MakePostFileData()
*  @brief     ƼƮ    κ
*  @return      Ʈ.
*  @param     CString & strBoundary
*  @param     CString & strAttachId
*  @param     CString & strTicket
*  @author	  yangjae.park(yangjae.park@nhn.com) 
*  @date      2012/02/23
*  @warning   
*/
CString UploadFile::MakePostFileData(const CString& strBoundary, LPCWSTR strAttachId, const CString& strTicket, LPCWSTR strUserID,  int nFileIndex)
{ 
	CString strFormat; 
	strFormat = _T("\r\n");
	strFormat += _T("--%s"); 
	strFormat += _T("\r\n"); 
	strFormat += _T("Content-Disposition: form-data; name=\"attachId\""); 
	strFormat += _T("\r\n\r\n"); 
	strFormat += _T("%s"); 
	strFormat += _T("\r\n"); 
	strFormat += _T("--%s"); 
	strFormat += _T("\r\n"); 
	strFormat += _T("Content-Disposition: form-data; name=\"ticket\""); 
	strFormat += _T("\r\n\r\n"); 
	strFormat += _T("%s"); 
	strFormat += _T("\r\n"); 
	strFormat += _T("--%s"); 
	strFormat += _T("\r\n"); 
	strFormat += _T("Content-Disposition: form-data; name=\"userid\""); 
	strFormat += _T("\r\n\r\n"); 
	strFormat += _T("%s"); 
	strFormat += _T("\r\n"); 
	strFormat += _T("--%s"); 
	strFormat += _T("\r\n"); 
	strFormat += _T("Content-Disposition: form-data; name=\"realIndex\""); 
	strFormat += _T("\r\n\r\n"); 
	strFormat += _T("%d"); 
	strFormat += _T("\r\n"); 
	strFormat += _T("--%s---"); 

	CString strData; 
	strData.Format(strFormat, strBoundary, strAttachId, strBoundary, strTicket, strBoundary, strUserID, strBoundary, nFileIndex, strBoundary); 

	return strData; 
} 


/*!
*  @fn CString MakeRequestHeaders()
*  @brief     ƼƮ   Boundary Header .
*  @return    Header Boundary
*  @param     CString & strBoundary
*  @author	  yangjae.park(yangjae.park@nhn.com) 
*  @date      2012/02/23
*  @warning   
*/
CString UploadFile::MakeRequestHeaders(const CString& strBoundary) 
{ 
	CString strFormat; 
	CString strData; 
	strFormat = _T("Content-Type: multipart/form-data; boundary=%s\r\n"); 
	strData.Format(strFormat, strBoundary); 
	return strData; 
} 

/*!
*  @fn void RequestPostData()
*  @brief     ƼƮ  ͸   κ
*  @return    
*  @param     CHttpFile * pHTTP 
*  @param     CString strFormData  
*  @author	  yangjae.park(yangjae.park@nhn.com) 
*  @date      2012/02/23
*  @warning   
*/
void UploadFile::RequestPostData(CHttpFile* pHTTP, CString strFormData)
{
	const int dwLen = WideCharToMultiByte(CP_UTF8, 0, strFormData, -1, NULL, 0, NULL, NULL)-1;//_tcslen(strFormData)+1;
	char *lpFormData = new char[dwLen];
	if (NULL != lpFormData)
	{
		WideCharToMultiByte(CP_UTF8, 0, strFormData, -1, lpFormData, dwLen, 0, 0); 
		pHTTP->Write((LPVOID)lpFormData, dwLen); 
		delete[] lpFormData;
	}
	return;
}

void ResponseWebData( CHttpFile * pFile, std::string &strResult) 
{
	const int BufferSize = 4096;
	char buffer[BufferSize];

	DWORD readCount;

	while(true)
	{
		readCount = pFile->Read(buffer, BufferSize);
		if (readCount <= 0) {
			break;
		}

		strResult += std::string(buffer, readCount);
	}
}

/*!
*  @fn BOOL FileSend()
*  @brief      MultiPart Ѵ
*  @return    TRUE  FALSE 
*  @param     CString strFilePath   ġ
*  @author	  yangjae.park(yangjae.park@nhn.com) 
*  @date      2012/02/23
*  @warning   
*/
BOOL UploadFile::FileSend(ThreadData *pData)
{
	CEvent *pEvent = pData->pStopEvent;
	CString strFilePath = pData->filePath;
	CWnd* pProgressWnd = pData->pProgressWnd;
	int fileAttachIndex = pData->fileAttachIndex;
	std::string fileAttachId = pData->fileAttachId;

	const wchar_t* strServerURL = L"http://api-nnote.ncloudbiz.com/api/note/attach/send.nhn";
	CString strServerName; 
	CString strObject; 
	INTERNET_PORT nPort; 
	DWORD dwServiceType; 

	TRACE("%s %d %d\n", __FUNCTION__, __LINE__, fileAttachIndex);
	CString strTicket;
	strTicket= String2CString(g_pMessengerActor->GetMessengerTicket());
	strTicket.Replace(_T("%3D%3D"), _T("=="));	// Send.nhn  ticket ڵؼ  ϴ ü  ϱ.  

	CFile file(strFilePath, CFile::modeRead | CFile::shareDenyNone);
	ULONGLONG dwFileLength = file.GetLength();
	DWORD dwReadLengthTotal = 0;

	AfxParseURL(strServerURL, dwServiceType, strServerName,  strObject, nPort); 

	ULONGLONG dwTotalRequestLength; 
	DWORD dwChunkLength; 
	DWORD dwReadLength; 
	CHttpFile* pHTTP = NULL; 
	dwChunkLength = 64 * 1024;  
	void* pBuffer = malloc(dwChunkLength); 
	if (NULL == pBuffer)
	{
		return FALSE;
	}

	CString strHTTPBoundary = _T("---------------------------7dc1736103ea");        
	CString strPreFileData = MakePreFileData(strHTTPBoundary, file.GetFileName()); 
	CString strPostFileData = MakePostFileData(strHTTPBoundary, Conversion::Utf8ToWide(fileAttachId).c_str(), strTicket, g_pMessengerActor->GetUserID(), fileAttachIndex); 
	CString strRequestHeaders = MakeRequestHeaders(strHTTPBoundary); 

	const int dwStrPreFileDataLen = WideCharToMultiByte(CP_UTF8, 0, strPreFileData, -1, NULL, 0, NULL, NULL) - 1;

	dwTotalRequestLength = dwStrPreFileDataLen + strPostFileData.GetLength() + dwFileLength; 
	
	CInternetSession session(_T("Ntalk/1.0 (PC Client)")); 
	CHttpConnection *connection = NULL; 	
	try
	{

		connection = session.GetHttpConnection(strServerName, nPort); 

		pHTTP = connection->OpenRequest(CHttpConnection::HTTP_VERB_POST, strObject);
		pHTTP->AddRequestHeaders(strRequestHeaders); 
		pHTTP->SendRequestEx((DWORD)dwTotalRequestLength, HSR_SYNC | HSR_INITIATE); 
		RequestPostData(pHTTP, strPreFileData);

		dwReadLength = -1; 
		ULONGLONG length = file.GetLength(); //used to calculate percentage complete. 
		while (0 != dwReadLength && ::WaitForSingleObject(pEvent->m_hObject, 0) == WAIT_TIMEOUT) 
		{ 
			dwReadLength = file.Read(pBuffer, dwChunkLength); 
			if (0 != dwReadLength) 
			{ 
				pHTTP->Write(pBuffer, dwReadLength);
				dwReadLengthTotal += dwReadLength;
				if(length!=dwReadLengthTotal)
				{
					if (pProgressWnd && pProgressWnd->GetSafeHwnd())
					{
						pProgressWnd->PostMessage(UM_MEMO_SEND_PROGRESS, (WPARAM)fileAttachIndex, (LPARAM)dwReadLengthTotal);
					}
				}

				Sleep(75);
			} 
		}

		free(pBuffer);
		file.Close(); 

		std::string strResult;
		if (length == dwReadLengthTotal)
		{
			// Ϸ ¿  ¸    Ѵ.
			if (pProgressWnd && pProgressWnd->GetSafeHwnd())
				pProgressWnd->PostMessage(UM_MEMO_FILESEND_RESULT_WAIT, (WPARAM)fileAttachIndex, 0);
	
			//Ƽ    ־ ٽ Ƽ Ȯؼ ش.
			strTicket= String2CString(g_pMessengerActor->GetMessengerTicket());
			strTicket.Replace(_T("%3D%3D"), _T("=="));
			strPostFileData = MakePostFileData(strHTTPBoundary, Conversion::Utf8ToWide(fileAttachId).c_str(), strTicket, g_pMessengerActor->GetUserID(), fileAttachIndex); 
			
			DWORD dwRet = 0;
			RequestPostData(pHTTP, strPostFileData);

			pHTTP->EndRequest(HSR_SYNC); 

			pHTTP->QueryInfoStatusCode(dwRet);
			ResponseWebData(pHTTP, strResult);
			if(dwRet == HTTP_STATUS_OK) 
				ResponseWebData(pHTTP, strResult);
			else
				NTALK_LOG->Error(L"[NOTE] ATTACH FILE EXCEPTION OCCURED %d:%d", fileAttachIndex, dwRet);

			// Ϸ    ҹư   ְ .
			if (pProgressWnd && pProgressWnd->GetSafeHwnd())
				pProgressWnd->PostMessage(UM_MEMO_FILESEND_RESULT_COMPLETE, (WPARAM)fileAttachIndex, 0);

			const FILE_UPLOAD_RESULT_DATA* fileUploadData = g_pMessengerActor->GetWASActor()->RequestUploadFileParsing(strResult);
			CString strCheckResult(fileUploadData->strResult.c_str());
			if (strCheckResult != L"OK")
			{
				if (pProgressWnd && pProgressWnd->GetSafeHwnd())
				{
					pProgressWnd->PostMessage(UM_MEMO_SEND_SERVER_ERROR, (WPARAM)fileUploadData, NULL);
					pProgressWnd->PostMessage(UM_MEMO_SEND_CANCEL, (WPARAM)fileAttachIndex, 0);
				}
			}

			else if (pProgressWnd && pProgressWnd->GetSafeHwnd())
			{
				pProgressWnd->PostMessage(UM_MEMO_SEND_PROGRESS, (WPARAM)fileAttachIndex, (LPARAM)dwReadLengthTotal);
			}
		}
	}
	catch (CInternetException* pEx) 
	{
		NTALK_LOG->Error(L"[NOTE] ATTACH FILE EXCEPTION OCCURED %d", fileAttachIndex);
		pProgressWnd->PostMessage(UM_MEMO_SEND_CANCEL, (WPARAM)fileAttachIndex, 0);
		pEx->Delete();
	}

	if (pHTTP != NULL)
		pHTTP->Close(); 
	
	delete pHTTP;
	connection->Close(); 
	delete connection;
	session.Close(); 

	return TRUE;
}
